import torch


def matmul_2x2_with_batched(left: torch.Tensor, right: torch.Tensor) -> torch.Tensor:
    result = torch.zeros_like(right)
    zero = torch.tensor(0, device=right.device)
    one = torch.tensor(1, device=right.device)
    result = result.index_add_(
        1,
        zero,
        right.select(1, 0).unsqueeze(1),
        alpha=left[0, 0],  # type: ignore [arg-type]
    )
    result = result.index_add_(
        1,
        zero,
        right.select(1, 1).unsqueeze(1),
        alpha=left[0, 1],  # type: ignore [arg-type]
    )
    result = result.index_add_(
        1,
        one,
        right.select(1, 0).unsqueeze(1),
        alpha=left[1, 0],  # type: ignore [arg-type]
    )
    result = result.index_add_(
        1,
        one,
        right.select(1, 1).unsqueeze(1),
        alpha=left[1, 1],  # type: ignore [arg-type]
    )
    return result



import torch
from typing import Callable

from emu_base.math.krylov_exp import DEFAULT_MAX_KRYLOV_DIM

max_krylov_dim = DEFAULT_MAX_KRYLOV_DIM


def double_krylov(
    op: Callable,
    state: torch.Tensor,
    grad: torch.Tensor,
    tolerance: float,
) -> tuple[list[torch.Tensor], torch.Tensor, list[torch.Tensor]]:
    """
    Returns a Lanczos decomposition of the Fréchet derivative of the exponential
    map U=exp(op) along the direction |state❭❬grad|.
    The decomposition is represented by the tuple (Vs, dS, Vg) such that,
        dU(op, |state❭❬grad|) = Vsᵗ @ dS @ Vg*

    Args:
        op (Callable): linear map to exponentiate, e.g. op(|ψ❭) = H|ψ❭.
        state (torch.Tensor):
        grad (torch.Tensor):
        tolerance (float): tolerance of the returned derivative.

    Returns:
        Vstate (list): Lanczos basis of state
        dS (torch.Tensor): matrix representing the derivative in the new basis
        Vgrad (list): Lanczos basis of grad

    Notes:
        Fréchet derivative dU(op,|a❭❬b|) being defined as:
            exp|op |a❭❬b|| = |exp(op) dU(op,|a❭❬b|)|
               |0   op   |   |0       exp(op)      |

        The function computes two Lanczos decomposition
        up to the given tolerance
            Va = Lanczos(|a❭,op(|a❭),op^2(|a❭),...)
            Vb = Lanczos(|b❭,op(|b❭),op^2(|b❭),...)
        such that,
            op = Vaᵗ @ Ta @ Va*
            op = Vbᵗ @ Tb @ Vb*

        In the new basis Va, Vb
            |op |a❭❬b|| -> |Ta  ab|0❭❬0||
            |0   op   |    |0   Tb      |
        where the top-right block only has one nonzero element.
        Exponentiating such matrix and selecting the top-right block
        gives us the desired matrix dS such that
            dU(op, |a❭❬b|) = Vaᵗ @ dS @ Vb*
    """
    Vs, Ts = lanczos(op, state, tolerance)
    Vg, Tg = lanczos(op, grad, tolerance)
    size_s = len(Vs)
    big_mat = torch.block_diag(Ts, Tg)
    # Only one element in the top-right corner
    big_mat[0, size_s] = state.norm() * grad.norm()
    dS = torch.matrix_exp(big_mat)[:size_s, size_s:]
    return Vs, dS, Vg


def lanczos(
    op: Callable,
    v: torch.Tensor,
    tolerance: float,
) -> tuple[list[torch.Tensor], torch.Tensor]:
    """
    Copy of the code in krylov_exp to do Laczos iteration
    To decide
     1. refactor this
     2. allow krylov results to store lanczos_vectors and T
    """
    converged = False
    lanczos_vectors = [v / v.norm()]
    T = torch.zeros(
        max_krylov_dim + 2, max_krylov_dim + 2, dtype=v.dtype, device=v.device
    )

    for j in range(max_krylov_dim):
        w = op(lanczos_vectors[-1])
        n = w.norm()
        for k in range(max(0, j - 1), j + 1):
            overlap = torch.tensordot(lanczos_vectors[k].conj(), w, dims=w.dim())
            T[k, j] = overlap
            w -= overlap * lanczos_vectors[k]

        n2 = w.norm()
        T[j + 1, j] = n2

        if n2 < tolerance:
            converged = True
            break

        lanczos_vectors.append(w / n2)
        # Compute exponential of extended T matrix
        T[j + 2, j + 1] = 1
        expd = torch.linalg.matrix_exp(T[: j + 3, : j + 3])

        # Local truncation error estimation
        err1 = abs(expd[j + 1, 0])
        err2 = abs(expd[j + 2, 0] * n)

        err = err1 if err1 < err2 else (err1 * err2 / (err1 - err2))
        if err < tolerance:
            converged = True
            break

    if not converged:
        raise RecursionError(
            "Lanczos iteration did not converge to precision in allotted number of steps."
        )
    size = len(lanczos_vectors)
    return lanczos_vectors, T[:size, :size]



import torch
from typing import Callable, Tuple

DEFAULT_MAX_KRYLOV_DIM: int = 200


def _lowest_eigen_pair(
    T_trunc: torch.Tensor,
) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Return the lowest eigenpair of the hermitian matrix T_trunc.
    """
    eig_energy, eig_state = torch.linalg.eigh(T_trunc)
    return eig_energy[0], eig_state[:, 0]


class KrylovEnergyResult:
    def __init__(
        self,
        ground_state: torch.Tensor,
        ground_energy: float,
        converged: bool,
        happy_breakdown: bool,
        iteration_count: int,
    ):
        self.ground_state = ground_state
        self.ground_energy = ground_energy
        self.converged = converged
        self.happy_breakdown = happy_breakdown
        self.iteration_count = iteration_count


def krylov_energy_minimization_impl(
    op: Callable[[torch.Tensor], torch.Tensor],
    psi_local: torch.Tensor,
    residual_tolerance: float,
    norm_tolerance: float,
    max_krylov_dim: int = DEFAULT_MAX_KRYLOV_DIM,
) -> KrylovEnergyResult:
    """
    Computes the ground state of a Hermitian operator using Lanczos algorithm.
    The Rayleigh quotient ⟨ψ|H|ψ⟩ is minimized over the Krylov subspace.

    The convergence of the results is determined by a residual norm criterion or a happy breakdown.
    """

    device = psi_local.device
    dtype = psi_local.dtype

    initial_norm = psi_local.norm()
    lanczos_vectors = [psi_local / initial_norm]
    T = torch.zeros(max_krylov_dim + 2, max_krylov_dim + 2, dtype=dtype, device=device)

    converged = False
    happy_breakdown = False
    iteration_count = 0

    for j in range(max_krylov_dim):
        w = op(lanczos_vectors[-1])

        for k in range(max(0, j - 1), j + 1):
            alpha = torch.tensordot(lanczos_vectors[k].conj(), w, dims=w.dim())
            T[k, j] = alpha
            w = w - alpha * lanczos_vectors[k]

        beta = w.norm()
        T[j + 1, j] = beta

        effective_dim = len(lanczos_vectors)
        size = effective_dim + (0 if beta < norm_tolerance else 1)
        T_truncated = T[:size, :size]

        ground_energy, ground_eigenvector = _lowest_eigen_pair(
            T_truncated
        )  # in Krylov subspace
        iteration_count = j + 1

        # happy breakdown check
        if beta < norm_tolerance:
            final_state = sum(
                c * vec for c, vec in zip(ground_eigenvector, lanczos_vectors)
            )
            final_state = final_state / final_state.norm()
            happy_breakdown = True
            converged = True
            break

        # Reconstruct final state in original Hilbert space
        lanczos_vectors.append(w / beta)
        final_state = sum(c * vec for c, vec in zip(ground_eigenvector, lanczos_vectors))
        final_state = final_state / final_state.norm()

        # residual norm convergence check
        residual_norm = torch.norm(op(final_state) - ground_energy * final_state)
        if residual_norm < residual_tolerance:
            happy_breakdown = False
            converged = True
            break

    return KrylovEnergyResult(
        ground_state=final_state,
        ground_energy=ground_energy.item(),
        converged=converged,
        happy_breakdown=happy_breakdown,
        iteration_count=iteration_count,
    )


def krylov_energy_minimization(
    op: Callable[[torch.Tensor], torch.Tensor],
    v: torch.Tensor,
    norm_tolerance: float,
    residual_tolerance: float,
    max_krylov_dim: int = DEFAULT_MAX_KRYLOV_DIM,
) -> Tuple[torch.Tensor, float]:

    result = krylov_energy_minimization_impl(
        op=op,
        psi_local=v,
        norm_tolerance=norm_tolerance,
        residual_tolerance=residual_tolerance,
        max_krylov_dim=max_krylov_dim,
    )

    if not result.converged and not result.happy_breakdown:
        raise RecursionError(
            "Krylov ground state solver did not converge within allotted iterations."
        )

    return result.ground_state, result.ground_energy



from pulser.noise_model import NoiseModel
import torch
import math

dtype = torch.complex128


def get_lindblad_operators(
    *,
    noise_type: str,
    noise_model: NoiseModel,
    interact_type: str = "ising",
    dim: int = 2,
) -> list[torch.Tensor]:

    assert noise_type in noise_model.noise_types

    if noise_type == "relaxation":
        c = math.sqrt(noise_model.relaxation_rate)
        relaxation = torch.zeros(dim, dim, dtype=dtype)
        relaxation[0, 1] = c
        return [relaxation]

    if noise_type == "dephasing":
        if noise_model.hyperfine_dephasing_rate != 0.0:
            raise NotImplementedError(
                "hyperfine_dephasing_rate is supported only in the digital basis"
            )

        c = math.sqrt(noise_model.dephasing_rate / 2)
        dephasing = torch.zeros(dim, dim, dtype=dtype)

        dephasing[0, 0] = c
        dephasing[1, 1] = -c

        return [dephasing]

    if noise_type == "depolarizing":
        c = math.sqrt(noise_model.depolarizing_rate / 4)
        depolarizing_x = torch.zeros(dim, dim, dtype=dtype)
        depolarizing_x[0, 1] = c
        depolarizing_x[1, 0] = c

        depolarizing_y = torch.zeros(dim, dim, dtype=dtype)
        depolarizing_y[0, 1] = torch.tensor(-c * 1.0j, dtype=dtype)
        depolarizing_y[1, 0] = torch.tensor(c * 1.0j, dtype=dtype)

        depolarizing_z = torch.zeros(dim, dim, dtype=dtype)
        depolarizing_z[0, 0] = c
        depolarizing_z[1, 1] = -c

        return [depolarizing_x, depolarizing_y, depolarizing_z]

    if noise_type == "eff_noise":
        if not all(
            isinstance(op, torch.Tensor) and op.shape == (dim, dim)
            for op in noise_model.eff_noise_opers
        ):
            raise ValueError(
                f"Only {dim} by {dim} effective noise operator matrices are "
                "supported and it should be given as torch tensors "
            )

        lindblad_ops = [  # lindblad operators with XY pulser basis are fine
            math.sqrt(rate) * torch.as_tensor(op)
            for rate, op in zip(noise_model.eff_noise_rates, noise_model.eff_noise_opers)
        ]

        # pulser ising basis changing to emu-mps ising basis
        if interact_type == "ising":
            for tensor in lindblad_ops:
                tensor[:2, :2] = torch.flip(tensor[:2, :2], (0, 1))

        return lindblad_ops
    if noise_type == "leakage":  # leakage operators are eff_noise
        return []

    raise ValueError(f"Unknown noise type: {noise_type}")


def compute_noise_from_lindbladians(
    lindbladians: list[torch.Tensor], dim: int = 2
) -> torch.Tensor:
    """
    Compute the single-qubit Hamiltonian noise term -0.5i∑L†L from all the
    given lindbladians.
    """
    assert all(
        lindbladian.shape == (dim, dim) for lindbladian in lindbladians
    ), "Only single-qubit lindblad operators are supported"

    zero = torch.zeros(dim, dim, dtype=dtype)

    return -0.5j * sum((L.mH @ L for L in lindbladians), start=zero)



from .constants import DEVICE_COUNT
from .pulser_adapter import PulserData, HamiltonianType
from .math.brents_root_finding import find_root_brents
from .math.krylov_exp import krylov_exp, DEFAULT_MAX_KRYLOV_DIM
from .jump_lindblad_operators import compute_noise_from_lindbladians
from .math.matmul import matmul_2x2_with_batched
from .utils import get_max_rss, apply_measurement_errors, unix_like, init_logging

__all__ = [
    "__version__",
    "get_max_rss",
    "compute_noise_from_lindbladians",
    "matmul_2x2_with_batched",
    "PulserData",
    "find_root_brents",
    "krylov_exp",
    "HamiltonianType",
    "DEFAULT_MAX_KRYLOV_DIM",
    "DEVICE_COUNT",
    "apply_measurement_errors",
    "unix_like",
    "init_logging",
]

__version__ = "2.5.0"



from typing import Sequence
from enum import Enum
import torch
import math
import pulser
from pulser.sampler import SequenceSamples
from pulser.noise_model import NoiseModel
from pulser.register.base_register import QubitId
from pulser.backend.config import EmulationConfig
from pulser._hamiltonian_data import HamiltonianData
from emu_base.jump_lindblad_operators import get_lindblad_operators


class HamiltonianType(Enum):
    Rydberg = 1
    XY = 2


_NON_LINDBLADIAN_NOISE = {"SPAM", "doppler", "amplitude", "detuning", "register"}


def _get_all_lindblad_noise_operators(
    noise_model: NoiseModel | None, dim: int = 2, interact_type: str = "ising"
) -> list[torch.Tensor]:
    if noise_model is None:
        return []

    return [
        op
        for noise_type in noise_model.noise_types
        if noise_type not in _NON_LINDBLADIAN_NOISE
        for op in get_lindblad_operators(
            noise_type=noise_type,
            noise_model=noise_model,
            dim=dim,
            interact_type=interact_type,
        )
    ]


def _get_target_times(
    sequence: pulser.Sequence, config: EmulationConfig, dt: int
) -> list[int]:
    sequence_duration = sequence.get_duration(include_fall_time=config.with_modulation)

    observable_times = set(range(0, sequence_duration, dt))
    observable_times.add(sequence_duration)
    for obs in config.observables:
        times: Sequence[float]
        if obs.evaluation_times is not None:
            times = obs.evaluation_times
        elif config.default_evaluation_times != "Full":
            times = config.default_evaluation_times.tolist()  # type: ignore[union-attr,assignment]
        observable_times |= set([round(time * sequence_duration) for time in times])

    target_times: list[int] = list(observable_times)
    target_times.sort()
    return target_times


def _extract_omega_delta_phi(
    noisy_samples: SequenceSamples,
    qubit_ids: tuple[str, ...],
    target_times: list[int],
) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    sequence_dict = noisy_samples.to_nested_dict(all_local=True, samples_type="tensor")[
        "Local"
    ]
    nsamples = len(target_times) - 1
    omega = torch.zeros(
        nsamples,
        len(qubit_ids),
        dtype=torch.complex128,
    )
    delta = torch.zeros(
        nsamples,
        len(qubit_ids),
        dtype=torch.complex128,
    )
    phi = torch.zeros(
        nsamples,
        len(qubit_ids),
        dtype=torch.complex128,
    )
    max_duration = noisy_samples.max_duration

    if "ground-rydberg" in sequence_dict and len(sequence_dict) == 1:
        locals_a_d_p = sequence_dict["ground-rydberg"]
    elif "XY" in sequence_dict and len(sequence_dict) == 1:
        locals_a_d_p = sequence_dict["XY"]
    else:
        raise ValueError("Only `ground-rydberg` and `mw_global` channels are supported.")
    for i in range(nsamples):
        t = (target_times[i] + target_times[i + 1]) / 2
        # The sampled values correspond to the start of each interval
        # To maximize the order of the solver, we need the values in the middle
        if math.ceil(t) < max_duration:
            # If we're not the final step, approximate this using linear
            # interpolation
            # Note that for dt even, t1=t2
            t1 = math.floor(t)
            t2 = math.ceil(t)
            for q_pos, q_id in enumerate(qubit_ids):
                omega[i, q_pos] = (
                    locals_a_d_p[q_id]["amp"][t1] + locals_a_d_p[q_id]["amp"][t2]
                ) / 2.0
                delta[i, q_pos] = (
                    locals_a_d_p[q_id]["det"][t1] + locals_a_d_p[q_id]["det"][t2]
                ) / 2.0
                phi[i, q_pos] = (
                    locals_a_d_p[q_id]["phase"][t1] + locals_a_d_p[q_id]["phase"][t2]
                ) / 2.0
        else:
            # We're in the final step and dt=1, approximate this using linear extrapolation
            # we can reuse omega_1 and omega_2 from before
            for q_pos, q_id in enumerate(qubit_ids):
                delta[i, q_pos] = (
                    3.0 * locals_a_d_p[q_id]["det"][t2] - locals_a_d_p[q_id]["det"][t1]
                ) / 2.0
                phi[i, q_pos] = (
                    3.0 * locals_a_d_p[q_id]["phase"][t2]
                    - locals_a_d_p[q_id]["phase"][t1]
                ) / 2.0
                omega[i, q_pos] = max(
                    (3.0 * locals_a_d_p[q_id]["amp"][t2] - locals_a_d_p[q_id]["amp"][t1])
                    / 2.0,
                    0.0,
                )
    return omega, delta, phi


class PulserData:
    slm_end_time: float
    full_interaction_matrix: torch.Tensor
    masked_interaction_matrix: torch.Tensor
    omega: torch.Tensor
    delta: torch.Tensor
    phi: torch.Tensor
    hamiltonian_type: HamiltonianType
    lindblad_ops: list[torch.Tensor]
    qubit_ids: tuple[QubitId, ...]

    def __init__(self, *, sequence: pulser.Sequence, config: EmulationConfig, dt: int):
        self.qubit_ids = sequence.register.qubit_ids
        self.qubit_count = len(self.qubit_ids)
        self.target_times = _get_target_times(sequence=sequence, config=config, dt=dt)
        self.hamiltonian = HamiltonianData.from_sequence(
            sequence,
            with_modulation=config.with_modulation,
            noise_model=config.noise_model,
        )

        self.omega, self.delta, self.phi = _extract_omega_delta_phi(
            self.hamiltonian.noisy_samples, self.qubit_ids, self.target_times
        )
        self.eigenstates = self.hamiltonian.eigenbasis

        int_type = self.hamiltonian.interaction_type
        self.dim = self.hamiltonian.dim
        if int_type == "ising":  # for local and global
            self.hamiltonian_type = HamiltonianType.Rydberg
        elif int_type == "XY":
            self.hamiltonian_type = HamiltonianType.XY
        else:
            raise ValueError(f"Unsupported basis: {int_type}")

        self.lindblad_ops = _get_all_lindblad_noise_operators(
            config.noise_model, dim=self.dim, interact_type=int_type
        )
        self.has_lindblad_noise: bool = self.lindblad_ops != []

        if config.interaction_matrix is not None:
            assert len(config.interaction_matrix) == self.qubit_count, (
                "The number of qubits in the register should be the same as the size of "
                "the interaction matrix"
            )

            self.full_interaction_matrix = config.interaction_matrix.as_tensor()
        else:
            self.full_interaction_matrix = (
                self.hamiltonian.noisy_interaction_matrix.as_tensor()
            )

        self.full_interaction_matrix[
            torch.abs(self.full_interaction_matrix) < config.interaction_cutoff
        ] = 0.0
        self.masked_interaction_matrix = self.full_interaction_matrix.clone()

        self.slm_end_time = (
            sequence._slm_mask_time[1] if len(sequence._slm_mask_time) > 1 else 0.0
        )

        # disable interaction for SLM masked qubits
        slm_targets = list(sequence._slm_mask_targets)
        for target in sequence.register.find_indices(slm_targets):
            self.masked_interaction_matrix[target] = 0.0
            self.masked_interaction_matrix[:, target] = 0.0



from collections import Counter
import random
import torch
import os
import logging
import sys
from pathlib import Path

unix_like = os.name != "nt"
if unix_like:
    from resource import RUSAGE_SELF, getrusage


def init_logging(log_level: int, log_file: Path | None) -> logging.Logger:
    logger = logging.getLogger("emulators")
    logger.propagate = False

    handler: logging.Handler
    if log_file is None:
        handler = logging.StreamHandler(sys.stdout)
    else:
        handler = logging.FileHandler(log_file, mode="w")

    handler.setLevel(log_level)
    handler.setFormatter(logging.Formatter("%(message)s"))
    for h in logger.handlers:
        logger.removeHandler(h)
    logger.addHandler(handler)
    return logger


def deallocate_tensor(t: torch.Tensor) -> None:
    """
    Free the memory used by a tensor. This is done regardless of the
    memory management done by Python: it is a forced deallocation
    that ignores the current reference count of the Tensor object.

    It is useful when you want to free memory that is no longer used
    inside a function but that memory is also owned by a variable
    in the outer scope, making it impossible to free it otherwise.

    After calling that function, the Tensor object
    should no longer be used.

    To work properly with e.g. tensordot but also user-created views,
    and since every view of a tensor owns the tensor's storage independently,
    it has to change the storage of the base AND every view referring to the base.
    However, it is not possible to access the views from the base, so
    if there are extra inaccessible views, it will raise an exception.
    """
    if (t._base is None and t._use_count() > 1) or (  # type: ignore[attr-defined]
        t._base is not None and t._base._use_count() > 2  # type: ignore[attr-defined]
    ):
        raise RuntimeError("Cannot deallocate tensor")

    replacement_storage = torch.zeros(0, dtype=t.dtype, device=t.device).untyped_storage()

    t.resize_(0)
    t.set_(source=replacement_storage)

    if t._base is not None:
        t._base.resize_(0)
        t._base.set_(source=replacement_storage)


def get_max_rss(gpu: bool) -> float:
    if gpu:
        max_mem_per_device = (
            torch.cuda.max_memory_allocated(device) * 1e-6
            for device in range(torch.cuda.device_count())
        )
        max_mem = max(max_mem_per_device)
    elif unix_like:
        max_mem = getrusage(RUSAGE_SELF).ru_maxrss * 1e-3
    else:
        return 0.0
    return max_mem


def readout_with_error(c: str, *, p_false_pos: float, p_false_neg: float) -> str:
    # p_false_pos = false positive, p_false_neg = false negative
    r = random.random()
    if c == "0" and r < p_false_pos:
        return "1"

    if c == "1" and r < p_false_neg:
        return "0"

    return c


def apply_measurement_errors(
    bitstrings: Counter[str], *, p_false_pos: float, p_false_neg: float
) -> Counter[str]:
    """
    Given a bag of sampled bitstrings, returns another bag of bitstrings
    sampled with readout/measurement errors.

        p_false_pos: probability of false positive
        p_false_neg: probability of false negative
    """

    result: Counter[str] = Counter()
    for bitstring, count in bitstrings.items():
        for _ in range(count):
            bitstring_with_error = "".join(
                readout_with_error(c, p_false_pos=p_false_pos, p_false_neg=p_false_neg)
                for c in bitstring
            )

            result[bitstring_with_error] += 1

    return result
